# Copyright (c) 2018-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import torch
import torch.nn.functional as F

from advertorch.utils import calc_l2distsq

from .base import Attack
from .base import LabelMixin


L2DIST_UPPER = 1e10
COEFF_UPPER = 1e10
INVALID_LABEL = -1
UPPER_CHECK = 1e9


class LBFGSAttack(Attack, LabelMixin):
    """
    The attack that uses L-BFGS to minimize the distance of the original
    and perturbed images

    :param predict: forward pass function.
    :param num_classes: number of clasess.
    :param batch_size: number of samples in the batch
    :param binary_search_steps: number of binary search times to find the
        optimum
    :param max_iterations: the maximum number of iterations
    :param initial_const: initial value of the constant c
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param loss_fn: loss function
    :param targeted: if the attack is targeted.
    """

    def __init__(self, predict, num_classes, batch_size=1,
                 binary_search_steps=9, max_iterations=100,
                 initial_const=1e-2,
                 clip_min=0, clip_max=1, loss_fn=None, targeted=False):
        super(LBFGSAttack, self).__init__(
            predict, loss_fn, clip_min, clip_max)
        # XXX: should combine the input loss function with other things
        self.num_classes = num_classes
        self.batch_size = batch_size
        self.binary_search_steps = binary_search_steps
        self.max_iterations = max_iterations
        self.initial_const = initial_const
        self.targeted = targeted


    def _update_if_better(
            self, adv_img, labs, output, dist, batch_size,
            final_l2dists, final_labels, final_advs):
        for ii in range(batch_size):
            target_label = labs[ii]
            output_logits = output[ii]
            _, output_label = torch.max(output_logits, 0)
            di = dist[ii]
            if (di < final_l2dists[ii] and
                    output_label.item() == target_label):
                final_l2dists[ii] = di
                final_labels[ii] = output_label
                final_advs[ii] = adv_img[ii]


    def _update_loss_coeffs(
            self, labs, batch_size,
            loss_coeffs, coeff_upper_bound, coeff_lower_bound, output):
        for ii in range(batch_size):
            _, cur_label = torch.max(output[ii], 0)
            if cur_label.item() == int(labs[ii]):
                coeff_upper_bound[ii] = min(
                    coeff_upper_bound[ii], loss_coeffs[ii])

                if coeff_upper_bound[ii] < UPPER_CHECK:
                    loss_coeffs[ii] = (
                        coeff_lower_bound[ii] + coeff_upper_bound[ii]) / 2
            else:
                coeff_lower_bound[ii] = max(
                    coeff_lower_bound[ii], loss_coeffs[ii])
                if coeff_upper_bound[ii] < UPPER_CHECK:
                    loss_coeffs[ii] = (
                        coeff_lower_bound[ii] + coeff_upper_bound[ii]) / 2
                else:
                    loss_coeffs[ii] *= 10


    def perturb(self, x, y=None):

        from scipy.optimize import fmin_l_bfgs_b

        def _loss_fn(adv_x_np, self, x, target, const):
            adv_x = torch.from_numpy(
                adv_x_np.reshape(x.shape)).float().to(
                x.device).requires_grad_()
            output = self.predict(adv_x)
            loss2 = torch.sum((x - adv_x) ** 2)
            loss_fn = F.cross_entropy(output, target, reduction='none')
            loss1 = torch.sum(const * loss_fn)
            loss = loss1 + loss2
            loss.backward()
            grad_ret = adv_x.grad.data.cpu().numpy().flatten().astype(float)
            loss = loss.data.cpu().numpy().flatten().astype(float)
            if not self.targeted:
                loss = -loss
            return loss, grad_ret

        x, y = self._verify_and_process_inputs(x, y)
        batch_size = len(x)
        coeff_lower_bound = x.new_zeros(batch_size)
        coeff_upper_bound = x.new_ones(batch_size) * COEFF_UPPER
        loss_coeffs = x.new_ones(batch_size) * self.initial_const
        final_l2dists = [L2DIST_UPPER] * batch_size
        final_labels = [INVALID_LABEL] * batch_size
        final_advs = x.clone()
        clip_min = self.clip_min * np.ones(x.shape[:]).astype(float)
        clip_max = self.clip_max * np.ones(x.shape[:]).astype(float)
        clip_bound = list(zip(clip_min.flatten(), clip_max.flatten()))

        for outer_step in range(self.binary_search_steps):
            init_guess = x.clone().cpu().numpy().flatten().astype(float)
            adv_x, f, _ = fmin_l_bfgs_b(_loss_fn,
                                        init_guess,
                                        args=(self, x.clone(), y, loss_coeffs),
                                        bounds=clip_bound,
                                        maxiter=self.max_iterations,
                                        iprint=0)

            adv_x = torch.from_numpy(
                adv_x.reshape(x.shape)).float().to(x.device)
            l2s = calc_l2distsq(x, adv_x)
            output = self.predict(adv_x)
            self._update_if_better(
                adv_x, y, output.data, l2s, batch_size,
                final_l2dists, final_labels, final_advs)
            self._update_loss_coeffs(
                y, batch_size,
                loss_coeffs, coeff_upper_bound, coeff_lower_bound,
                output.data)
        return final_advs



# Copyright (c) 2018-present, Royal Bank of Canada and other authors.
# See the AUTHORS.txt file for a list of contributors.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import numpy as np
import torch
import torch.nn as nn

from advertorch.utils import clamp
from advertorch.utils import normalize_by_pnorm
from advertorch.utils import clamp_by_pnorm
from advertorch.utils import is_float_or_torch_tensor
from advertorch.utils import batch_multiply
from advertorch.utils import batch_clamp
from advertorch.utils import replicate_input
from advertorch.utils import batch_l1_proj

from .base import Attack
from .base import LabelMixin
from .utils import rand_init_delta


def perturb_iterative(xvar, yvar, predict, nb_iter, eps, eps_iter, loss_fn,
                      delta_init=None, minimize=False, ord=np.inf,
                      clip_min=0.0, clip_max=1.0,
                      l1_sparsity=None):
    """
    Iteratively maximize the loss over the input. It is a shared method for
    iterative attacks including IterativeGradientSign, LinfPGD, etc.

    :param xvar: input data.
    :param yvar: input labels.
    :param predict: forward pass function.
    :param nb_iter: number of iterations.
    :param eps: maximum distortion.
    :param eps_iter: attack step size.
    :param loss_fn: loss function.
    :param delta_init: (optional) tensor contains the random initialization.
    :param minimize: (optional bool) whether to minimize or maximize the loss.
    :param ord: (optional) the order of maximum distortion (inf or 2).
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param l1_sparsity: sparsity value for L1 projection.
                  - if None, then perform regular L1 projection.
                  - if float value, then perform sparse L1 descent from
                    Algorithm 1 in https://arxiv.org/pdf/1904.13000v1.pdf
    :return: tensor containing the perturbed input.
    """
    if delta_init is not None:
        delta = delta_init
    else:
        delta = torch.zeros_like(xvar)

    delta.requires_grad_()
    for ii in range(nb_iter):
        outputs = predict(xvar + delta)
        loss = loss_fn(outputs, yvar)
        if minimize:
            loss = -loss

        loss.backward()
        if ord == np.inf:
            grad_sign = delta.grad.data.sign()
            delta.data = delta.data + batch_multiply(eps_iter, grad_sign)
            delta.data = batch_clamp(eps, delta.data)
            delta.data = clamp(xvar.data + delta.data, clip_min, clip_max
                               ) - xvar.data

        elif ord == 2:
            grad = delta.grad.data
            grad = normalize_by_pnorm(grad)
            delta.data = delta.data + batch_multiply(eps_iter, grad)
            delta.data = clamp(xvar.data + delta.data, clip_min, clip_max
                               ) - xvar.data
            if eps is not None:
                delta.data = clamp_by_pnorm(delta.data, ord, eps)

        elif ord == 1:
            grad = delta.grad.data
            abs_grad = torch.abs(grad)

            batch_size = grad.size(0)
            view = abs_grad.view(batch_size, -1)
            view_size = view.size(1)
            if l1_sparsity is None:
                vals, idx = view.topk(1)
            else:
                vals, idx = view.topk(
                    int(np.round((1 - l1_sparsity) * view_size)))

            out = torch.zeros_like(view).scatter_(1, idx, vals)
            out = out.view_as(grad)
            grad = grad.sign() * (out > 0).float()
            grad = normalize_by_pnorm(grad, p=1)
            delta.data = delta.data + batch_multiply(eps_iter, grad)

            delta.data = batch_l1_proj(delta.data.cpu(), eps)
            if xvar.is_cuda:
                delta.data = delta.data.cuda()
            delta.data = clamp(xvar.data + delta.data, clip_min, clip_max
                               ) - xvar.data
        else:
            error = "Only ord = inf, ord = 1 and ord = 2 have been implemented"
            raise NotImplementedError(error)
        delta.grad.data.zero_()

    x_adv = clamp(xvar + delta, clip_min, clip_max)
    return x_adv


class PGDAttack(Attack, LabelMixin):
    """
    The projected gradient descent attack (Madry et al, 2017).
    The attack performs nb_iter steps of size eps_iter, while always staying
    within eps from the initial point.
    Paper: https://arxiv.org/pdf/1706.06083.pdf

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations.
    :param eps_iter: attack step size.
    :param rand_init: (optional bool) random initialization.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param ord: (optional) the order of maximum distortion (inf or 2).
    :param targeted: if the attack is targeted.
    """

    def __init__(
            self, predict, loss_fn=None, eps=0.3, nb_iter=40,
            eps_iter=0.01, rand_init=True, clip_min=0., clip_max=1.,
            ord=np.inf, l1_sparsity=None, targeted=False):
        """
        Create an instance of the PGDAttack.

        """
        super(PGDAttack, self).__init__(
            predict, loss_fn, clip_min, clip_max)
        self.eps = eps
        self.nb_iter = nb_iter
        self.eps_iter = eps_iter
        self.rand_init = rand_init
        self.ord = ord
        self.targeted = targeted
        if self.loss_fn is None:
            self.loss_fn = nn.CrossEntropyLoss(reduction="sum")
        self.l1_sparsity = l1_sparsity
        assert is_float_or_torch_tensor(self.eps_iter)
        assert is_float_or_torch_tensor(self.eps)

    def perturb(self, x, y=None):
        """
        Given examples (x, y), returns their adversarial counterparts with
        an attack length of eps.

        :param x: input tensor.
        :param y: label tensor.
                  - if None and self.targeted=False, compute y as predicted
                    labels.
                  - if self.targeted=True, then y must be the targeted labels.
        :return: tensor containing perturbed inputs.
        """
        x, y = self._verify_and_process_inputs(x, y)

        delta = torch.zeros_like(x)
        delta = nn.Parameter(delta)
        if self.rand_init:
            rand_init_delta(
                delta, x, self.ord, self.eps, self.clip_min, self.clip_max)
            delta.data = clamp(
                x + delta.data, min=self.clip_min, max=self.clip_max) - x

        rval = perturb_iterative(
            x, y, self.predict, nb_iter=self.nb_iter,
            eps=self.eps, eps_iter=self.eps_iter,
            loss_fn=self.loss_fn, minimize=self.targeted,
            ord=self.ord, clip_min=self.clip_min,
            clip_max=self.clip_max, delta_init=delta,
            l1_sparsity=self.l1_sparsity,
        )

        return rval.data


class LinfPGDAttack(PGDAttack):
    """
    PGD Attack with order=Linf

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations.
    :param eps_iter: attack step size.
    :param rand_init: (optional bool) random initialization.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: if the attack is targeted.
    """

    def __init__(
            self, predict, loss_fn=None, eps=0.3, nb_iter=40,
            eps_iter=0.01, rand_init=True, clip_min=0., clip_max=1.,
            targeted=False):
        ord = np.inf
        super(LinfPGDAttack, self).__init__(
            predict=predict, loss_fn=loss_fn, eps=eps, nb_iter=nb_iter,
            eps_iter=eps_iter, rand_init=rand_init, clip_min=clip_min,
            clip_max=clip_max, targeted=targeted,
            ord=ord)


class L2PGDAttack(PGDAttack):
    """
    PGD Attack with order=L2

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations.
    :param eps_iter: attack step size.
    :param rand_init: (optional bool) random initialization.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: if the attack is targeted.
    """

    def __init__(
            self, predict, loss_fn=None, eps=0.3, nb_iter=40,
            eps_iter=0.01, rand_init=True, clip_min=0., clip_max=1.,
            targeted=False):
        ord = 2
        super(L2PGDAttack, self).__init__(
            predict=predict, loss_fn=loss_fn, eps=eps, nb_iter=nb_iter,
            eps_iter=eps_iter, rand_init=rand_init, clip_min=clip_min,
            clip_max=clip_max, targeted=targeted,
            ord=ord)


class L1PGDAttack(PGDAttack):
    """
    PGD Attack with order=L1

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations.
    :param eps_iter: attack step size.
    :param rand_init: (optional bool) random initialization.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: if the attack is targeted.
    """

    def __init__(
            self, predict, loss_fn=None, eps=10., nb_iter=40,
            eps_iter=0.01, rand_init=True, clip_min=0., clip_max=1.,
            targeted=False):
        ord = 1
        super(L1PGDAttack, self).__init__(
            predict=predict, loss_fn=loss_fn, eps=eps, nb_iter=nb_iter,
            eps_iter=eps_iter, rand_init=rand_init, clip_min=clip_min,
            clip_max=clip_max, targeted=targeted,
            ord=ord, l1_sparsity=None)


class SparseL1DescentAttack(PGDAttack):
    """
    SparseL1Descent Attack

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations.
    :param eps_iter: attack step size.
    :param rand_init: (optional bool) random initialization.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: if the attack is targeted.
    :param l1_sparsity: proportion of zeros in gradient updates
    """

    def __init__(
            self, predict, loss_fn=None, eps=0.3, nb_iter=40,
            eps_iter=0.01, rand_init=False, clip_min=0., clip_max=1.,
            l1_sparsity=0.95, targeted=False):
        ord = 1
        super(SparseL1DescentAttack, self).__init__(
            predict=predict, loss_fn=loss_fn, eps=eps, nb_iter=nb_iter,
            eps_iter=eps_iter, rand_init=rand_init, clip_min=clip_min,
            clip_max=clip_max, targeted=targeted,
            ord=ord, l1_sparsity=l1_sparsity)


class L2BasicIterativeAttack(PGDAttack):
    """Like GradientAttack but with several steps for each epsilon.

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations.
    :param eps_iter: attack step size.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: if the attack is targeted.
    """

    def __init__(self, predict, loss_fn=None, eps=0.1, nb_iter=10,
                 eps_iter=0.05, clip_min=0., clip_max=1., targeted=False):
        ord = 2
        rand_init = False
        l1_sparsity = None
        super(L2BasicIterativeAttack, self).__init__(
            predict, loss_fn, eps, nb_iter, eps_iter, rand_init,
            clip_min, clip_max, ord, l1_sparsity, targeted)


class LinfBasicIterativeAttack(PGDAttack):
    """
    Like GradientSignAttack but with several steps for each epsilon.
    Aka Basic Iterative Attack.
    Paper: https://arxiv.org/pdf/1611.01236.pdf

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations.
    :param eps_iter: attack step size.
    :param rand_init: (optional bool) random initialization.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: if the attack is targeted.
    """

    def __init__(self, predict, loss_fn=None, eps=0.1, nb_iter=10,
                 eps_iter=0.05, clip_min=0., clip_max=1., targeted=False):
        ord = np.inf
        rand_init = False
        l1_sparsity = None
        super(LinfBasicIterativeAttack, self).__init__(
            predict, loss_fn, eps, nb_iter, eps_iter, rand_init,
            clip_min, clip_max, ord, l1_sparsity, targeted)


class MomentumIterativeAttack(Attack, LabelMixin):
    """
    The Momentum Iterative Attack (Dong et al. 2017).

    The attack performs nb_iter steps of size eps_iter, while always staying
    within eps from the initial point. The optimization is performed with
    momentum.
    Paper: https://arxiv.org/pdf/1710.06081.pdf

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations
    :param decay_factor: momentum decay factor.
    :param eps_iter: attack step size.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: if the attack is targeted.
    :param ord: the order of maximum distortion (inf or 2).
    """

    def __init__(
            self, predict, loss_fn=None, eps=0.3, nb_iter=40, decay_factor=1.,
            eps_iter=0.01, clip_min=0., clip_max=1., targeted=False,
            ord=np.inf):
        """Create an instance of the MomentumIterativeAttack."""
        super(MomentumIterativeAttack, self).__init__(
            predict, loss_fn, clip_min, clip_max)
        self.eps = eps
        self.nb_iter = nb_iter
        self.decay_factor = decay_factor
        self.eps_iter = eps_iter
        self.targeted = targeted
        self.ord = ord
        if self.loss_fn is None:
            self.loss_fn = nn.CrossEntropyLoss(reduction="sum")

    def perturb(self, x, y=None):
        """
        Given examples (x, y), returns their adversarial counterparts with
        an attack length of eps.

        :param x: input tensor.
        :param y: label tensor.
                  - if None and self.targeted=False, compute y as predicted
                    labels.
                  - if self.targeted=True, then y must be the targeted labels.
        :return: tensor containing perturbed inputs.
        """
        x, y = self._verify_and_process_inputs(x, y)

        delta = torch.zeros_like(x)
        g = torch.zeros_like(x)

        delta = nn.Parameter(delta)

        for i in range(self.nb_iter):

            if delta.grad is not None:
                delta.grad.detach_()
                delta.grad.zero_()

            imgadv = x + delta
            outputs = self.predict(imgadv)
            loss = self.loss_fn(outputs, y)
            if self.targeted:
                loss = -loss
            loss.backward()

            g = self.decay_factor * g + normalize_by_pnorm(
                delta.grad.data, p=1)
            # according to the paper it should be .sum(), but in their
            #   implementations (both cleverhans and the link from the paper)
            #   it is .mean(), but actually it shouldn't matter
            if self.ord == np.inf:
                delta.data += batch_multiply(self.eps_iter, torch.sign(g))
                delta.data = batch_clamp(self.eps, delta.data)
                delta.data = clamp(
                    x + delta.data, min=self.clip_min, max=self.clip_max) - x
            elif self.ord == 2:
                delta.data += self.eps_iter * normalize_by_pnorm(g, p=2)
                delta.data *= clamp(
                    (self.eps * normalize_by_pnorm(delta.data, p=2) /
                        delta.data),
                    max=1.)
                delta.data = clamp(
                    x + delta.data, min=self.clip_min, max=self.clip_max) - x
            else:
                error = "Only ord = inf and ord = 2 have been implemented"
                raise NotImplementedError(error)

        rval = x + delta.data
        return rval


class L2MomentumIterativeAttack(MomentumIterativeAttack):
    """
    The L2 Momentum Iterative Attack
    Paper: https://arxiv.org/pdf/1710.06081.pdf

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations
    :param decay_factor: momentum decay factor.
    :param eps_iter: attack step size.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: if the attack is targeted.
    """

    def __init__(
            self, predict, loss_fn=None, eps=0.3, nb_iter=40, decay_factor=1.,
            eps_iter=0.01, clip_min=0., clip_max=1., targeted=False):
        """Create an instance of the MomentumIterativeAttack."""
        ord = 2
        super(L2MomentumIterativeAttack, self).__init__(
            predict, loss_fn, eps, nb_iter, decay_factor,
            eps_iter, clip_min, clip_max, targeted, ord)


class LinfMomentumIterativeAttack(MomentumIterativeAttack):
    """
    The Linf Momentum Iterative Attack
    Paper: https://arxiv.org/pdf/1710.06081.pdf

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param nb_iter: number of iterations
    :param decay_factor: momentum decay factor.
    :param eps_iter: attack step size.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: if the attack is targeted.
    """

    def __init__(
            self, predict, loss_fn=None, eps=0.3, nb_iter=40, decay_factor=1.,
            eps_iter=0.01, clip_min=0., clip_max=1., targeted=False):
        """Create an instance of the MomentumIterativeAttack."""
        ord = np.inf
        super(LinfMomentumIterativeAttack, self).__init__(
            predict, loss_fn, eps, nb_iter, decay_factor,
            eps_iter, clip_min, clip_max, targeted, ord)


class FastFeatureAttack(Attack):
    """
    Fast attack against a target internal representation of a model using
    gradient descent (Sabour et al. 2016).
    Paper: https://arxiv.org/abs/1511.05122

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: maximum distortion.
    :param eps_iter: attack step size.
    :param nb_iter: number of iterations
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    """

    def __init__(self, predict, loss_fn=None, eps=0.3, eps_iter=0.05,
                 nb_iter=10, rand_init=True, clip_min=0., clip_max=1.):
        """Create an instance of the FastFeatureAttack."""
        super(FastFeatureAttack, self).__init__(
            predict, loss_fn, clip_min, clip_max)
        self.eps = eps
        self.eps_iter = eps_iter
        self.nb_iter = nb_iter
        self.rand_init = rand_init
        self.clip_min = clip_min
        self.clip_max = clip_max
        if self.loss_fn is None:
            self.loss_fn = nn.MSELoss(reduction="sum")

    def perturb(self, source, guide, delta=None):
        """
        Given source, returns their adversarial counterparts
        with representations close to that of the guide.

        :param source: input tensor which we want to perturb.
        :param guide: targeted input.
        :param delta: tensor contains the random initialization.
        :return: tensor containing perturbed inputs.
        """
        # Initialization
        if delta is None:
            delta = torch.zeros_like(source)
            if self.rand_init:
                delta = delta.uniform_(-self.eps, self.eps)
        else:
            delta = delta.detach()

        delta.requires_grad_()

        source = replicate_input(source)
        guide = replicate_input(guide)
        guide_ftr = self.predict(guide).detach()

        xadv = perturb_iterative(source, guide_ftr, self.predict,
                                 self.nb_iter, eps_iter=self.eps_iter,
                                 loss_fn=self.loss_fn, minimize=True,
                                 ord=np.inf, eps=self.eps,
                                 clip_min=self.clip_min,
                                 clip_max=self.clip_max,
                                 delta_init=delta)

        xadv = clamp(xadv, self.clip_min, self.clip_max)

        return xadv.data



# Copyright (c) 2018-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from advertorch.functional import FloatToIntSqueezing

from .base import Processor


class BitSqueezing(Processor):
    """
    Bit Squeezing.

    :param bit_depth: bit depth.
    :param vmin: min value.
    :param vmax: max value.
    """

    def __init__(self, bit_depth, vmin=0., vmax=1.):
        super(BitSqueezing, self).__init__()

        self.bit_depth = bit_depth
        self.max_int = 2 ** self.bit_depth - 1
        self.vmin = vmin
        self.vmax = vmax


    def forward(self, x):
        return FloatToIntSqueezing.apply(
            x, self.max_int, self.vmin, self.vmax)


class BinaryFilter(BitSqueezing):
    """
    Binary Filter.

    :param vmin: min value.
    :param vmax: max value.
    """

    def __init__(self, vmin=0., vmax=1.):
        super(BinaryFilter, self).__init__(bit_depth=1, vmin=vmin, vmax=vmax)



# Copyright (c) 2018-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from contextlib import contextmanager


class ctx_noparamgrad(object):
    def __init__(self, module):
        self.prev_grad_state = get_param_grad_state(module)
        self.module = module
        set_param_grad_off(module)

    def __enter__(self):
        pass

    def __exit__(self, *args):
        set_param_grad_state(self.module, self.prev_grad_state)
        return False


class ctx_eval(object):
    def __init__(self, module):
        self.prev_training_state = get_module_training_state(module)
        self.module = module
        set_module_training_off(module)

    def __enter__(self):
        pass

    def __exit__(self, *args):
        set_module_training_state(self.module, self.prev_training_state)
        return False


@contextmanager
def ctx_noparamgrad_and_eval(module):
    with ctx_noparamgrad(module) as a, ctx_eval(module) as b:
        yield (a, b)


def get_module_training_state(module):
    return {mod: mod.training for mod in module.modules()}


def set_module_training_state(module, training_state):
    for mod in module.modules():
        mod.training = training_state[mod]


def set_module_training_off(module):
    for mod in module.modules():
        mod.training = False


def get_param_grad_state(module):
    return {param: param.requires_grad for param in module.parameters()}


def set_param_grad_state(module, grad_state):
    for param in module.parameters():
        param.requires_grad = grad_state[param]


def set_param_grad_off(module):
    for param in module.parameters():
        param.requires_grad = False



# Copyright (c) 2018-present, Royal Bank of Canada and other authors.
# See the AUTHORS.txt file for a list of contributors.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

# flake8: noqa

from .base import Attack
from .base import LabelMixin

from .one_step_gradient import GradientAttack
from .one_step_gradient import GradientSignAttack
from .one_step_gradient import FGM
from .one_step_gradient import FGSM

from .iterative_projected_gradient import FastFeatureAttack
from .iterative_projected_gradient import L2BasicIterativeAttack
from .iterative_projected_gradient import LinfBasicIterativeAttack
from .iterative_projected_gradient import PGDAttack
from .iterative_projected_gradient import LinfPGDAttack
from .iterative_projected_gradient import L2PGDAttack
from .iterative_projected_gradient import L1PGDAttack
from .iterative_projected_gradient import SparseL1DescentAttack
from .iterative_projected_gradient import MomentumIterativeAttack
from .iterative_projected_gradient import L2MomentumIterativeAttack
from .iterative_projected_gradient import LinfMomentumIterativeAttack

from .carlini_wagner import CarliniWagnerL2Attack
from .ead import ElasticNetL1Attack

from .decoupled_direction_norm import DDNL2Attack

from .lbfgs import LBFGSAttack

from .localsearch import SinglePixelAttack
from .localsearch import LocalSearchAttack

from .spatial import SpatialTransformAttack

from .jsma import JacobianSaliencyMapAttack
from .jsma import JSMA

from .spsa import LinfSPSAAttack
from .fast_adaptive_boundary import FABAttack
from .fast_adaptive_boundary import LinfFABAttack
from .fast_adaptive_boundary import L2FABAttack
from .fast_adaptive_boundary import L1FABAttack

from .utils import ChooseBestAttack



# Copyright (c) 2018-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import pathlib
from torch.utils.data.dataset import Subset

import numpy as np
import torch
import torchvision.transforms as transforms
import torchvision.datasets as datasets

from advertorch.test_utils import LeNet5

# TODO: need to refactor path to keep a single copy of file

ROOT_PATH = os.path.expanduser("~/.advertorch")
DATA_PATH = os.path.join(ROOT_PATH, "data")

path_of_this_module = os.path.dirname(sys.modules[__name__].__file__)
TRAINED_MODEL_PATH = os.path.join(path_of_this_module, "trained_models")


def mkdir(directory):
    pathlib.Path(directory).mkdir(parents=True, exist_ok=True)


def get_mnist_train_loader(batch_size, shuffle=True):
    loader = torch.utils.data.DataLoader(
        datasets.MNIST(DATA_PATH, train=True, download=True,
                       transform=transforms.ToTensor()),
        batch_size=batch_size, shuffle=shuffle)
    loader.name = "mnist_train"
    return loader


def get_mnist_test_loader(batch_size, shuffle=False):
    loader = torch.utils.data.DataLoader(
        datasets.MNIST(DATA_PATH, train=False, download=True,
                       transform=transforms.ToTensor()),
        batch_size=batch_size, shuffle=shuffle)
    loader.name = "mnist_test"
    return loader


def get_cifar10_train_loader(batch_size, shuffle=True):
    loader = torch.utils.data.DataLoader(
        datasets.CIFAR10(DATA_PATH, train=True, download=True,
                         transform=transforms.ToTensor()),
        batch_size=batch_size, shuffle=shuffle)
    loader.name = "cifar10_train"
    return loader


def get_cifar10_test_loader(batch_size, shuffle=False):
    loader = torch.utils.data.DataLoader(
        datasets.CIFAR10(DATA_PATH, train=False, download=True,
                         transform=transforms.ToTensor()),
        batch_size=batch_size, shuffle=shuffle)
    loader.name = "cifar10_test"
    return loader


def get_mnist_lenet5_clntrained():
    filename = "mnist_lenet5_clntrained.pt"
    model = LeNet5()
    model.load_state_dict(
        torch.load(os.path.join(TRAINED_MODEL_PATH, filename)))
    model.eval()
    model.name = "MNIST LeNet5 standard training"
    # TODO: also described where can you find this model, and how is it trained
    return model


def get_mnist_lenet5_advtrained():
    filename = "mnist_lenet5_advtrained.pt"
    model = LeNet5()
    model.load_state_dict(
        torch.load(os.path.join(TRAINED_MODEL_PATH, filename)))
    model.eval()
    model.name = "MNIST LeNet 5 PGD training according to Madry et al. 2018"
    # TODO: also described where can you find this model, and how is it trained
    return model


def get_madry_et_al_cifar10_train_transform():
    return transforms.Compose([
        transforms.Pad(4, padding_mode="reflect"),
        transforms.RandomCrop(32),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
    ])



def get_train_val_loaders(
        dataset, datapath=DATA_PATH,
        train_size=None, val_size=5000,
        train_batch_size=100, val_batch_size=1000,
        kwargs=None, train_transform=None, val_transform=None,
        train_shuffle=True, val_shuffle=False):
    """Support MNIST and CIFAR10"""
    if kwargs is None:
        kwargs = {}
    if train_transform is None:
        train_transform = transforms.ToTensor()
    if val_transform is None:
        val_transform = transforms.ToTensor()

    datapath = os.path.join(datapath, dataset)

    trainset = datasets.__dict__[dataset](
        datapath, train=True, download=True, transform=train_transform)

    if train_size is not None:
        assert train_size + val_size <= len(trainset)

    if val_size > 0:
        indices = list(range(len(trainset)))
        trainset = Subset(trainset, indices[val_size:])

        valset = datasets.__dict__[dataset](
            datapath, train=True, download=True, transform=val_transform)
        valset = Subset(valset, indices[:val_size])
        val_loader = torch.utils.data.DataLoader(
            valset, batch_size=val_batch_size, shuffle=val_shuffle, **kwargs)

    else:
        val_loader = None

    if train_size is not None:
        trainset = Subset(trainset, list(range(train_size)))

    train_loader = torch.utils.data.DataLoader(
        trainset, batch_size=train_batch_size, shuffle=train_shuffle, **kwargs)

    return train_loader, val_loader


def get_test_loader(
        dataset, datapath=DATA_PATH, test_size=None, batch_size=1000,
        transform=None, kwargs=None, shuffle=False):
    """Support MNIST and CIFAR10"""
    if kwargs is None:
        kwargs = {}
    if transform is None:
        transform = transforms.ToTensor()

    datapath = os.path.join(datapath, dataset)

    testset = datasets.__dict__[dataset](
        datapath, train=False, download=True, transform=transform)

    if test_size is not None:
        testset = Subset(testset, list(range(test_size)))

    test_loader = torch.utils.data.DataLoader(
        testset, batch_size=batch_size, shuffle=shuffle, **kwargs)
    return test_loader


def bchw2bhwc(x):
    if isinstance(x, np.ndarray):
        pass
    else:
        raise

    if x.ndim == 3:
        return np.moveaxis(x, 0, 2)
    if x.ndim == 4:
        return np.moveaxis(x, 1, 3)


def bhwc2bchw(x):
    if isinstance(x, np.ndarray):
        pass
    else:
        raise

    if x.ndim == 3:
        return np.moveaxis(x, 2, 0)
    if x.ndim == 4:
        return np.moveaxis(x, 3, 1)


def _imshow(img):
    import matplotlib.pyplot as plt
    img = bchw2bhwc(img.detach().cpu().numpy())
    if img.shape[2] == 1:
        img = np.repeat(img, 3, axis=2)
    plt.imshow(img, vmin=0, vmax=1)
    plt.axis("off")


class ImageNetClassNameLookup(object):

    def _load_list(self):
        import json
        with open(self.json_path) as f:
            class_idx = json.load(f)
        self.label2classname = [
            class_idx[str(k)][1] for k in range(len(class_idx))]

    def __init__(self):
        self.json_url = ("https://s3.amazonaws.com/deep-learning-models/"
                         "image-models/imagenet_class_index.json")
        self.json_path = os.path.join(DATA_PATH, "imagenet_class_index.json")
        if os.path.exists(self.json_path):
            self._load_list()
        else:
            import urllib
            urllib.request.urlretrieve(self.json_url, self.json_path)
            self._load_list()


    def __call__(self, label):
        return self.label2classname[label]


def get_panda_image():
    img_path = os.path.join(DATA_PATH, "panda.jpg")
    img_url = "https://farm1.static.flickr.com/230/524562325_fb0a11d1e1.jpg"

    def _load_panda_image():
        from skimage.io import imread
        return imread(img_path) / 255.

    if os.path.exists(img_path):
        return _load_panda_image()
    else:
        import urllib
        urllib.request.urlretrieve(img_url, img_path)
        return _load_panda_image()


mkdir(ROOT_PATH)
mkdir(DATA_PATH)



import torch



x = torch.tensor([1., 2.])
x.requires_grad_()
y = x ** 2

# print(torch.autograd.grad(y, x, torch.tensor([1., 0.]), retain_graph=True))
# print(torch.autograd.grad(y, x, torch.tensor([0., 1.]), retain_graph=True))



# Copyright (c) 2018-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import numpy as np
from scipy import ndimage

from advertorch.test_utils import generate_data_model_on_img
from advertorch.utils import torch_allclose
from advertorch.defenses import BinaryFilter
from advertorch.defenses import MedianSmoothing2D

data, label, model = generate_data_model_on_img()


def test_binary_filter():
    assert torch_allclose(BinaryFilter()(data), data > 0.5)


def test_median_filter():
    # XXX: doesn't pass when kernel_size is even
    # XXX: when kernel_size is odd, pixels on the boundaries are different
    kernel_size = 3
    padding = kernel_size // 2
    rval_scipy = ndimage.filters.median_filter(
        data.detach().numpy(), size=(1, 1, kernel_size, kernel_size))
    rval = MedianSmoothing2D(kernel_size=kernel_size)(data).detach().numpy()
    assert np.allclose(rval_scipy[:, :, padding:-padding, padding:-padding],
                       rval[:, :, padding:-padding, padding:-padding])


# TODO: correctness test of GaussianSmoothing2D and AverageSmoothing2D

if __name__ == '__main__':
    test_binary_filter()
    test_median_filter()



import math
import torch.nn as nn
import torch.nn.functional as F


class LeNet5Madry(nn.Module):
    # model replicated from
    #   https://github.com/MadryLab/mnist_challenge/blob/
    #   2527d24c4c34e511a12b8a9d7cf6b949aae6fc1b/model.py
    # TODO: combine with the model in advertorch.test_utils

    def __init__(
            self, nb_filters=(1, 32, 64), kernel_sizes=(5, 5),
            paddings=(2, 2), strides=(1, 1), pool_sizes=(2, 2),
            nb_hiddens=(7 * 7 * 64, 1024), nb_classes=10):
        super(LeNet5Madry, self).__init__()
        self.conv1 = nn.Conv2d(
            nb_filters[0], nb_filters[1], kernel_size=kernel_sizes[0],
            padding=paddings[0], stride=strides[0])
        self.relu1 = nn.ReLU(inplace=True)
        self.maxpool1 = nn.MaxPool2d(pool_sizes[0])
        self.conv2 = nn.Conv2d(
            nb_filters[1], nb_filters[2], kernel_size=kernel_sizes[1],
            padding=paddings[0], stride=strides[0])
        self.relu2 = nn.ReLU(inplace=True)
        self.maxpool2 = nn.MaxPool2d(pool_sizes[1])
        self.linear1 = nn.Linear(nb_hiddens[0], nb_hiddens[1])
        self.relu3 = nn.ReLU(inplace=True)
        self.linear2 = nn.Linear(nb_hiddens[1], nb_classes)

    def forward(self, x):
        out = self.maxpool1(self.relu1(self.conv1(x)))
        out = self.maxpool2(self.relu2(self.conv2(out)))
        out = out.view(out.size(0), -1)
        out = self.relu3(self.linear1(out))
        out = self.linear2(out)
        return out


def get_lenet5madry_with_width(widen_factor):
    return LeNet5Madry(
        nb_filters=(1, int(widen_factor * 32), int(widen_factor * 64)),
        nb_hiddens=(7 * 7 * int(widen_factor * 64), int(widen_factor * 1024)))


# WideResNet related code adapted from
#   https://github.com/xternalz/WideResNet-pytorch/blob/
#   ae12d25bdf273010bd4a54971948a6c796cb95ed/wideresnet.py


class BasicBlock(nn.Module):
    def __init__(self, in_planes, out_planes, stride, drop_rate=0.0):
        super(BasicBlock, self).__init__()
        self.bn1 = nn.BatchNorm2d(in_planes)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv1 = nn.Conv2d(
            in_planes, out_planes, kernel_size=3, stride=stride,
            padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(out_planes)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(
            out_planes, out_planes, kernel_size=3, stride=1,
            padding=1, bias=False)
        self.drop_rate = drop_rate
        self.in_out_equal = (in_planes == out_planes)

        if not self.in_out_equal:
            self.conv_shortcut = nn.Conv2d(
                in_planes, out_planes, kernel_size=1, stride=stride,
                padding=0, bias=False)

    def forward(self, x):
        out = self.relu1(self.bn1(x))
        if not self.in_out_equal:
            x = self.conv_shortcut(out)
        out = self.relu2(self.bn2(self.conv1(out)))
        if self.drop_rate > 0:
            out = F.dropout(out, p=self.drop_rate, training=self.training)
        out = self.conv2(out)
        out += x
        return out


class ConvGroup(nn.Module):
    def __init__(
            self, num_blocks, in_planes, out_planes, block, stride,
            drop_rate=0.0):
        super(ConvGroup, self).__init__()
        self.layer = self._make_layer(
            block, in_planes, out_planes, num_blocks, stride, drop_rate)

    def _make_layer(
            self, block, in_planes, out_planes, num_blocks, stride, drop_rate):
        layers = []
        for i in range(int(num_blocks)):
            layers.append(
                block(in_planes=in_planes if i == 0 else out_planes,
                      out_planes=out_planes,
                      stride=stride if i == 0 else 1,
                      drop_rate=drop_rate)
            )
        return nn.Sequential(*layers)

    def forward(self, x):
        return self.layer(x)


class WideResNet(nn.Module):
    def __init__(self, depth, num_classes, widen_factor=1, drop_rate=0.0,
                 color_channels=3, block=BasicBlock):
        super(WideResNet, self).__init__()
        num_channels = [
            16, int(16 * widen_factor),
            int(32 * widen_factor), int(64 * widen_factor)]
        assert((depth - 4) % 6 == 0)
        num_blocks = (depth - 4) / 6

        self.conv1 = nn.Conv2d(
            color_channels, num_channels[0], kernel_size=3, stride=1,
            padding=1, bias=False)
        self.convgroup1 = ConvGroup(
            num_blocks, num_channels[0], num_channels[1], block, 1, drop_rate)
        self.convgroup2 = ConvGroup(
            num_blocks, num_channels[1], num_channels[2], block, 2, drop_rate)
        self.convgroup3 = ConvGroup(
            num_blocks, num_channels[2], num_channels[3], block, 2, drop_rate)
        # global average pooling and classifier
        self.bn1 = nn.BatchNorm2d(num_channels[3])
        self.relu = nn.ReLU(inplace=True)
        self.fc = nn.Linear(num_channels[3], num_classes)
        self.num_channels = num_channels[3]

        for mod in self.modules():
            if isinstance(mod, nn.Conv2d):
                n = mod.kernel_size[0] * mod.kernel_size[1] * mod.out_channels
                mod.weight.data.normal_(0, math.sqrt(2. / n))
            elif isinstance(mod, nn.BatchNorm2d):
                mod.weight.data.fill_(1)
                mod.bias.data.zero_()
            elif isinstance(mod, nn.Linear):
                mod.bias.data.zero_()

    def forward(self, x):
        out = self.conv1(x)
        out = self.convgroup1(out)
        out = self.convgroup2(out)
        out = self.convgroup3(out)
        out = self.relu(self.bn1(out))
        out = out.mean(dim=-1).mean(dim=-1)
        out = self.fc(out)
        return out


def get_cifar10_wrn28_widen_factor(widen_factor):
    from advertorch.utils import PerImageStandardize
    model = WideResNet(28, 10, widen_factor)
    model = nn.Sequential(PerImageStandardize(), model)
    return model



# Copyright (c) 2018-present, Royal Bank of Canada and other authors.
# See the AUTHORS.txt file for a list of contributors.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import warnings
import pytest
import random

import numpy as np
import torch
import torch.nn.parallel
import torch.optim
import torch.utils.data
import tensorflow as tf

from cleverhans.attacks import CarliniWagnerL2
from cleverhans.attacks import ElasticNetMethod
from cleverhans.attacks import FastGradientMethod
from cleverhans.attacks import MomentumIterativeMethod
from cleverhans.attacks import MadryEtAl
from cleverhans.attacks import FastFeatureAdversaries
from cleverhans.attacks import BasicIterativeMethod
from cleverhans.attacks import LBFGS
from cleverhans.attacks import SaliencyMapMethod
from cleverhans.model import Model as ClModel

from advertorch.attacks import CarliniWagnerL2Attack
from advertorch.attacks import ElasticNetL1Attack
from advertorch.attacks import GradientAttack
from advertorch.attacks import GradientSignAttack
from advertorch.attacks import L2MomentumIterativeAttack
from advertorch.attacks import LinfMomentumIterativeAttack
from advertorch.attacks import LinfPGDAttack
from advertorch.attacks import FastFeatureAttack
from advertorch.attacks import LinfBasicIterativeAttack
from advertorch.attacks import L2BasicIterativeAttack
from advertorch.attacks import LBFGSAttack
from advertorch.attacks import JacobianSaliencyMapAttack
from advertorch.test_utils import SimpleModel
from advertorch.test_utils import merge2dicts


BATCH_SIZE = 9
DIM_INPUT = 15
NUM_CLASS = 5
EPS = 0.08

ATOL = 1e-4
RTOL = 1e-4
NB_ITER = 5

# XXX: carlini still doesn't pass sometimes under certain random seed
seed = 66666
torch.manual_seed(seed)
np.random.seed(seed)
random.seed(seed)
tf.set_random_seed(seed)
inputs = np.random.uniform(0, 1, size=(BATCH_SIZE, DIM_INPUT))
targets = np.random.randint(0, NUM_CLASS, size=BATCH_SIZE)


targets_onehot = np.zeros((BATCH_SIZE, NUM_CLASS), dtype='int')
targets_onehot[np.arange(BATCH_SIZE), targets] = 1


class SimpleModelTf(ClModel):

    def __init__(self, dim_input, num_classes, session=None):
        import keras
        self.sess = session
        model = keras.models.Sequential()
        model.add(keras.layers.Dense(10, input_shape=(dim_input, )))
        model.add(keras.layers.Activation('relu'))
        model.add(keras.layers.Dense(num_classes))
        self.model = model
        self.flag_weight_set = False

    def set_weights(self, weights):
        self.model.set_weights(weights)
        self.flag_weight_set = True

    def load_state_dict(self, w):
        self.set_weights([
            w['fc1.weight'].cpu().numpy().transpose(),
            w['fc1.bias'].cpu().numpy(),
            w['fc2.weight'].cpu().numpy().transpose(),
            w['fc2.bias'].cpu().numpy(),
        ])

    def get_logits(self, data):
        assert self.flag_weight_set, "Weight Not Set!!!"
        return self.model(data)

    def get_probs(self, data):
        assert self.flag_weight_set, "Weight Not Set!!!"
        return tf.nn.softmax(logits=self.model(data))


def load_weights_pt(model_pt, layers):
    w = model_pt.state_dict()
    layers[0].W = tf.Variable(tf.convert_to_tensor(
        w['fc1.weight'].cpu().numpy().transpose(), tf.float32))
    layers[0].b = tf.Variable(tf.convert_to_tensor(
        w['fc1.bias'].cpu().numpy(), tf.float32))
    layers[2].W = tf.Variable(tf.convert_to_tensor(
        w['fc2.weight'].cpu().numpy().transpose(), tf.float32))
    layers[2].b = tf.Variable(tf.convert_to_tensor(
        w['fc2.bias'].cpu().numpy(), tf.float32))


def setup_simple_model_tf(model_pt, input_shape):
    from cleverhans_tutorials.tutorial_models import MLP, Linear, ReLU
    layers = [Linear(10),
              ReLU(),
              Linear(10)]
    layers[0].name = 'fc1'
    layers[1].name = 'relu'
    layers[2].name = 'fc2'
    model = MLP(layers, input_shape)
    load_weights_pt(model_pt, layers)
    return model



# kwargs for attacks to be tested
attack_kwargs = {
    GradientSignAttack: {
        "cl_class": FastGradientMethod,
        "kwargs": dict(
            eps=EPS,
            clip_min=0.0,
            clip_max=1.0,
        ),
        "at_kwargs": dict(
        ),
        "cl_kwargs": dict(
            ord=np.inf,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    GradientAttack: {
        "cl_class": FastGradientMethod,
        "kwargs": dict(
            eps=EPS,
            clip_min=0.0,
            clip_max=1.0,
        ),
        "at_kwargs": dict(
        ),
        "cl_kwargs": dict(
            ord=2,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    LinfPGDAttack: {
        "cl_class": MadryEtAl,
        "kwargs": dict(
            eps=EPS,
            eps_iter=0.01,
            clip_min=0.0,
            clip_max=1.0,
            rand_init=False,
            nb_iter=NB_ITER,
        ),
        "at_kwargs": dict(
            targeted=True,
        ),
        "cl_kwargs": dict(
            ord=np.inf,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    L2MomentumIterativeAttack: {
        "cl_class": MomentumIterativeMethod,
        "kwargs": dict(
            eps=EPS,
            eps_iter=0.01,
            clip_min=0.0,
            clip_max=1.0,
            decay_factor=1.,
            nb_iter=NB_ITER,
        ),
        "at_kwargs": dict(
        ),
        "cl_kwargs": dict(
            ord=2,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    LinfMomentumIterativeAttack: {
        "cl_class": MomentumIterativeMethod,
        "kwargs": dict(
            eps=EPS,
            eps_iter=0.01,
            clip_min=0.0,
            clip_max=1.0,
            decay_factor=1.,
            nb_iter=NB_ITER,
        ),
        "at_kwargs": dict(
        ),
        "cl_kwargs": dict(
            ord=np.inf,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    CarliniWagnerL2Attack: {
        "cl_class": CarliniWagnerL2,
        "kwargs": dict(
            max_iterations=100,
            clip_min=0,
            clip_max=1,
            binary_search_steps=9,
            learning_rate=0.1,
            confidence=0.1,
        ),
        "at_kwargs": dict(
            num_classes=NUM_CLASS,
        ),
        "cl_kwargs": dict(
            batch_size=BATCH_SIZE,
            initial_const=1e-3,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    ElasticNetL1Attack: {
        "cl_class": ElasticNetMethod,
        "kwargs": dict(
            max_iterations=100,
            clip_min=0,
            clip_max=1,
            binary_search_steps=9,
            learning_rate=0.1,
            confidence=0.1,
        ),
        "at_kwargs": dict(
            num_classes=NUM_CLASS,
        ),
        "cl_kwargs": dict(
            batch_size=BATCH_SIZE,
            initial_const=1e-3,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    FastFeatureAttack: {
        "cl_class": FastFeatureAdversaries,
        "kwargs": dict(
            nb_iter=NB_ITER,
            clip_min=0,
            clip_max=1,
            eps_iter=0.05,
            eps=0.3,
        ),
        "at_kwargs": dict(
        ),
        "cl_kwargs": dict(
            layer='logits',
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    LinfBasicIterativeAttack: {
        "cl_class": BasicIterativeMethod,
        "kwargs": dict(
            clip_min=0,
            clip_max=1,
            eps_iter=0.05,
            eps=0.1,
            nb_iter=NB_ITER,
        ),
        "at_kwargs": dict(
        ),
        "cl_kwargs": dict(
            ord=np.inf,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    L2BasicIterativeAttack: {
        "cl_class": BasicIterativeMethod,
        "kwargs": dict(
            clip_min=0,
            clip_max=1,
            eps_iter=0.05,
            eps=0.1,
            nb_iter=NB_ITER,
        ),
        "at_kwargs": dict(
        ),
        "cl_kwargs": dict(
            ord=2,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    LBFGSAttack: {
        "cl_class": LBFGS,
        "kwargs": dict(
            clip_min=0.,
            clip_max=1.,
            # set binary search step = 3, which can successfully create
            # adversarial images and the difference between advertorch
            # and cleverhans is within the threshold
            # the difference of the two results are very small at first
            # because of some rounding and calculating difference
            # with tensors and numpy arrays, the difference gets larger
            # with more iterations
            binary_search_steps=3,
            max_iterations=50,
            initial_const=1e-3,
            batch_size=BATCH_SIZE,
        ),
        "at_kwargs": dict(
            num_classes=NUM_CLASS,
        ),
        "cl_kwargs": dict(
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
    JacobianSaliencyMapAttack: {
        "cl_class": SaliencyMapMethod,
        "kwargs": dict(
            clip_min=0.0,
            clip_max=1.0,
            theta=1.0,
            gamma=1.0,
        ),
        "at_kwargs": dict(
            num_classes=NUM_CLASS,
            comply_cleverhans=True,
        ),
        "cl_kwargs": dict(
            # nb_classes=NUM_CLASS,
        ),
        "thresholds": dict(
            atol=ATOL,
            rtol=RTOL,
        ),
    },
}


def overwrite_fastfeature(attack, x, g, eta, **kwargs):
    # overwrite cleverhans generate function for fastfeatureattack to
    # allow eta as an input
    from cleverhans.utils_tf import clip_eta

    # Parse and save attack-specific parameters
    assert attack.parse_params(**kwargs)

    g_feat = attack.model.get_layer(g, attack.layer)

    # Initialize loop variables
    eta = tf.Variable(tf.convert_to_tensor(eta, np.float32))
    eta = clip_eta(eta, attack.ord, attack.eps)

    for i in range(attack.nb_iter):
        eta = attack.attack_single_step(x, eta, g_feat)

    # Define adversarial example (and clip if necessary)
    adv_x = x + eta
    if attack.clip_min is not None and attack.clip_max is not None:
        adv_x = tf.clip_by_value(adv_x, attack.clip_min, attack.clip_max)

    return adv_x


def genenerate_ptb_pt(adversary, inputs, targets, delta=None):
    if inputs.ndim == 4:
        # TODO: move the transpose to a better place
        input_t = torch.from_numpy(inputs.transpose(0, 3, 1, 2))
    else:
        input_t = torch.from_numpy(inputs)
    input_t = input_t.float()

    if targets is None:
        adversary.targeted = False
        adv_pt = adversary.perturb(input_t, None)
    else:
        target_t = torch.from_numpy(targets)
        if isinstance(adversary, FastFeatureAttack):
            adv_pt = adversary.perturb(input_t, target_t,
                                       delta=torch.from_numpy(delta))
        else:
            adversary.targeted = True
            adv_pt = adversary.perturb(input_t, target_t)

    adv_pt = adv_pt.cpu().detach().numpy()

    if inputs.ndim == 4:
        # TODO: move the transpose to a better place
        adv_pt = adv_pt.transpose(0, 2, 3, 1)
    return adv_pt - inputs


def compare_at_cl(ptb_at, ptb_cl, atol, rtol):
    assert np.allclose(ptb_at, ptb_cl, atol=atol, rtol=rtol), \
        (np.abs(ptb_at - ptb_cl).max())


def compare_attacks(key, item, targeted=False):
    AdvertorchAttack = key
    CleverhansAttack = item["cl_class"]
    cl_kwargs = merge2dicts(item["kwargs"], item["cl_kwargs"])
    at_kwargs = merge2dicts(item["kwargs"], item["at_kwargs"])
    thresholds = item["thresholds"]
    seed = 6666
    torch.manual_seed(seed)
    np.random.seed(seed)

    # WARNING: don't use tf.InteractiveSession() here
    # It causes that fastfeature attack has to be the last test for some reason
    with tf.Session() as sess:
        model_pt = SimpleModel(DIM_INPUT, NUM_CLASS)
        model_tf = SimpleModelTf(DIM_INPUT, NUM_CLASS)
        model_tf.load_state_dict(model_pt.state_dict())
        adversary = AdvertorchAttack(model_pt, **at_kwargs)

        if AdvertorchAttack is FastFeatureAttack:
            model_tf_fastfeature = setup_simple_model_tf(
                model_pt, inputs.shape)
            delta = np.random.uniform(
                -item["kwargs"]['eps'], item["kwargs"]['eps'],
                size=inputs.shape).astype('float32')
            inputs_guide = np.random.uniform(
                0, 1, size=(BATCH_SIZE, DIM_INPUT)).astype('float32')
            inputs_tf = tf.convert_to_tensor(inputs, np.float32)
            inputs_guide_tf = tf.convert_to_tensor(inputs_guide, np.float32)
            attack = CleverhansAttack(model_tf_fastfeature)
            cl_result = overwrite_fastfeature(attack,
                                              x=inputs_tf,
                                              g=inputs_guide_tf,
                                              eta=delta,
                                              **cl_kwargs)
            init = tf.global_variables_initializer()
            sess.run(init)
            ptb_cl = sess.run(cl_result) - inputs
            ptb_at = genenerate_ptb_pt(
                adversary, inputs, inputs_guide, delta=delta)

        else:
            attack = CleverhansAttack(model_tf, sess=sess)
            if targeted:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    ptb_cl = attack.generate_np(
                        inputs, y_target=targets_onehot, **cl_kwargs) - inputs
                ptb_at = genenerate_ptb_pt(adversary, inputs, targets=targets)
            else:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    ptb_cl = attack.generate_np(
                        inputs, y=None, **cl_kwargs) - inputs
                ptb_at = genenerate_ptb_pt(adversary, inputs, targets=None)

        if AdvertorchAttack is CarliniWagnerL2Attack:
            assert np.sum(np.abs(ptb_at)) > 0 and np.sum(np.abs(ptb_cl)) > 0, \
                ("Both advertorch and cleverhans returns zero perturbation"
                 " of CarliniWagnerL2Attack, "
                 "the test results are not reliable,"
                 " Adjust your testing parameters to avoid this."
                 )
        compare_at_cl(ptb_at, ptb_cl, **thresholds)


@pytest.mark.parametrize("targeted", [False, True])
def test_fgsm_attack(targeted):
    compare_attacks(
        GradientSignAttack,
        attack_kwargs[GradientSignAttack],
        targeted)


@pytest.mark.parametrize("targeted", [False, True])
def test_fgm_attack(targeted):
    compare_attacks(
        GradientAttack,
        attack_kwargs[GradientAttack],
        targeted)


@pytest.mark.parametrize("targeted", [False, True])
def test_l2_momentum_iterative_attack(targeted):
    compare_attacks(
        L2MomentumIterativeAttack,
        attack_kwargs[L2MomentumIterativeAttack],
        targeted)


@pytest.mark.parametrize("targeted", [False, True])
def test_linf_momentum_iterative_attack(targeted):
    compare_attacks(
        LinfMomentumIterativeAttack,
        attack_kwargs[LinfMomentumIterativeAttack],
        targeted)


@pytest.mark.skip(reason="XXX: temporary")
def test_fastfeature_attack():
    compare_attacks(
        FastFeatureAttack,
        attack_kwargs[FastFeatureAttack])


@pytest.mark.parametrize("targeted", [False, True])
def test_pgd_attack(targeted):
    compare_attacks(
        LinfPGDAttack,
        attack_kwargs[LinfPGDAttack],
        targeted)


@pytest.mark.parametrize("targeted", [False, True])
def test_iterative_sign_attack(targeted):
    compare_attacks(
        LinfBasicIterativeAttack,
        attack_kwargs[LinfBasicIterativeAttack],
        targeted)


@pytest.mark.parametrize("targeted", [False, True])
def test_iterative_attack(targeted):
    compare_attacks(
        L2BasicIterativeAttack,
        attack_kwargs[L2BasicIterativeAttack],
        targeted)


@pytest.mark.parametrize("targeted", [False, True])
def test_carlini_l2_attack(targeted):
    compare_attacks(
        CarliniWagnerL2Attack,
        attack_kwargs[CarliniWagnerL2Attack],
        targeted)


@pytest.mark.parametrize("targeted", [False, True])
def test_elasticnet_l1_attack(targeted):
    compare_attacks(
        ElasticNetL1Attack,
        attack_kwargs[ElasticNetL1Attack],
        targeted)


def test_lbfgs_attack():
    compare_attacks(
        LBFGSAttack,
        attack_kwargs[LBFGSAttack],
        True)


@pytest.mark.skip(reason="XXX: temporary")
def test_jsma():
    compare_attacks(
        JacobianSaliencyMapAttack,
        attack_kwargs[JacobianSaliencyMapAttack],
        True)


if __name__ == '__main__':
    # pass
    test_iterative_attack(False)
    test_iterative_attack(True)



# Copyright (c) 2018-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from advertorch.functional import JPEGEncodingDecoding

from .base import Processor


class JPEGFilter(Processor):
    """
    JPEG Filter.

    :param quality: quality of the output.
    """
    def __init__(self, quality=75):
        super(JPEGFilter, self).__init__()
        self.quality = quality

    def forward(self, x):
        return JPEGEncodingDecoding.apply(x, self.quality)



# Copyright (c) 2018-present, Royal Bank of Canada and other authors.
# See the AUTHORS.txt file for a list of contributors.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import os
from setuptools import setup
from setuptools import find_packages


with open(os.path.join(os.path.dirname(__file__), 'advertorch/VERSION')) as f:
    version = f.read().strip()


setup(name='advertorch',
      version=version,
      url='https://github.com/BorealisAI/advertorch',
      package_data={'advertorch_examples': ['*.ipynb', 'trained_models/*.pt']},
      install_requires=[],
      include_package_data=True,
      packages=find_packages())



# Copyright (c) 2018-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import torch.nn as nn

from advertorch.utils import clamp
from advertorch.utils import normalize_by_pnorm
from advertorch.utils import batch_multiply

from .base import Attack
from .base import LabelMixin


class GradientSignAttack(Attack, LabelMixin):
    """
    One step fast gradient sign method (Goodfellow et al, 2014).
    Paper: https://arxiv.org/abs/1412.6572

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: attack step size.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: indicate if this is a targeted attack.
    """

    def __init__(self, predict, loss_fn=None, eps=0.3, clip_min=0.,
                 clip_max=1., targeted=False):
        """
        Create an instance of the GradientSignAttack.
        """
        super(GradientSignAttack, self).__init__(
            predict, loss_fn, clip_min, clip_max)

        self.eps = eps
        self.targeted = targeted
        if self.loss_fn is None:
            self.loss_fn = nn.CrossEntropyLoss(reduction="sum")

    def perturb(self, x, y=None):
        """
        Given examples (x, y), returns their adversarial counterparts with
        an attack length of eps.

        :param x: input tensor.
        :param y: label tensor.
                  - if None and self.targeted=False, compute y as predicted
                    labels.
                  - if self.targeted=True, then y must be the targeted labels.
        :return: tensor containing perturbed inputs.
        """

        x, y = self._verify_and_process_inputs(x, y)
        xadv = x.requires_grad_()
        outputs = self.predict(xadv)

        loss = self.loss_fn(outputs, y)
        if self.targeted:
            loss = -loss
        loss.backward()
        grad_sign = xadv.grad.detach().sign()

        xadv = xadv + batch_multiply(self.eps, grad_sign)

        xadv = clamp(xadv, self.clip_min, self.clip_max)

        return xadv.detach()


FGSM = GradientSignAttack


class GradientAttack(Attack, LabelMixin):
    """
    Perturbs the input with gradient (not gradient sign) of the loss wrt the
    input.

    :param predict: forward pass function.
    :param loss_fn: loss function.
    :param eps: attack step size.
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param targeted: indicate if this is a targeted attack.
    """

    def __init__(self, predict, loss_fn=None, eps=0.3,
                 clip_min=0., clip_max=1., targeted=False):
        """
        Create an instance of the GradientAttack.
        """
        super(GradientAttack, self).__init__(
            predict, loss_fn, clip_min, clip_max)

        self.eps = eps
        self.targeted = targeted
        if self.loss_fn is None:
            self.loss_fn = nn.CrossEntropyLoss(reduction="sum")

    def perturb(self, x, y=None):
        """
        Given examples (x, y), returns their adversarial counterparts with
        an attack length of eps.

        :param x: input tensor.
        :param y: label tensor.
                  - if None and self.targeted=False, compute y as predicted
                    labels.
                  - if self.targeted=True, then y must be the targeted labels.
        :return: tensor containing perturbed inputs.
        """
        x, y = self._verify_and_process_inputs(x, y)
        xadv = x.requires_grad_()
        outputs = self.predict(xadv)

        loss = self.loss_fn(outputs, y)
        if self.targeted:
            loss = -loss
        loss.backward()
        grad = normalize_by_pnorm(xadv.grad)
        xadv = xadv + batch_multiply(self.eps, grad)
        xadv = clamp(xadv, self.clip_min, self.clip_max)

        return xadv.detach()


FGM = GradientAttack



# Copyright (c) 2018-present, Royal Bank of Canada and other authors.
# See the AUTHORS.txt file for a list of contributors.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import os
import sys

import torch
import torchvision

import advertorch
from advertorch.attacks.utils import multiple_mini_batch_attack


def get_benchmark_sys_info():
    rval = "#\n#\n"
    rval += ("# Automatically generated benchmark report "
             "(screen print of running this file)\n#\n")
    uname = os.uname()
    rval += "# sysname: {}\n".format(uname.sysname)
    rval += "# release: {}\n".format(uname.release)
    rval += "# version: {}\n".format(uname.version)
    rval += "# machine: {}\n".format(uname.machine)
    rval += "# python: {}.{}.{}\n".format(
        sys.version_info.major,
        sys.version_info.minor,
        sys.version_info.micro)
    rval += "# torch: {}\n".format(torch.__version__)
    rval += "# torchvision: {}\n".format(torchvision.__version__)
    rval += "# advertorch: {}\n".format(advertorch.__version__)
    return rval


def _calculate_benchmark_results(
        model, loader, attack_class, attack_kwargs, norm, device, num_batch):
    adversary = attack_class(model, **attack_kwargs)
    label, pred, advpred, dist = multiple_mini_batch_attack(
        adversary, loader, device=device, norm=norm, num_batch=num_batch)
    accuracy = 100. * (label == pred).sum().item() / len(label)
    attack_success_rate = 100. * (label != advpred).sum().item() / len(label)
    dist = None if dist is None else dist[(label != advpred) & (label == pred)]
    return len(label), accuracy, attack_success_rate, dist


def _generate_basic_benchmark_str(
        model, loader, attack_class, attack_kwargs, num, accuracy,
        attack_success_rate):
    rval = ""
    rval += "# attack type: {}\n".format(attack_class.__name__)

    prefix = " attack kwargs: "
    count = 0
    for key in attack_kwargs:
        this_prefix = prefix if count == 0 else " " * len(prefix)
        count += 1
        rval += "#{}{}={}\n".format(this_prefix, key, attack_kwargs[key])

    rval += "# data: {}, {} samples\n".format(loader.name, num)
    rval += "# model: {}\n".format(model.name)
    rval += "# accuracy: {}%\n".format(accuracy)
    rval += "# attack success rate: {}%\n".format(attack_success_rate)
    return rval


def benchmark_attack_success_rate(
        model, loader, attack_class, attack_kwargs,
        device="cuda", num_batch=None):
    num, accuracy, attack_success_rate, _ = _calculate_benchmark_results(
        model, loader, attack_class, attack_kwargs, None, device, num_batch)
    rval = _generate_basic_benchmark_str(
        model, loader, attack_class, attack_kwargs, num, accuracy,
        attack_success_rate)
    return rval


def benchmark_margin(
        model, loader, attack_class, attack_kwargs, norm,
        device="cuda", num_batch=None):

    num, accuracy, attack_success_rate, dist = _calculate_benchmark_results(
        model, loader, attack_class, attack_kwargs, norm, device, num_batch)
    rval = _generate_basic_benchmark_str(
        model, loader, attack_class, attack_kwargs, num, accuracy,
        attack_success_rate)

    rval += "# Among successful attacks ({} norm) ".format(norm) + \
        "on correctly classified examples:\n"
    rval += "#    minimum distance: {:.4}\n".format(dist.min().item())
    rval += "#    median distance: {:.4}\n".format(dist.median().item())
    rval += "#    maximum distance: {:.4}\n".format(dist.max().item())
    rval += "#    average distance: {:.4}\n".format(dist.mean().item())
    rval += "#    distance standard deviation: {:.4}\n".format(
        dist.std().item())

    return rval



# Copyright (c) 2018-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import torch
import torch.nn as nn
import torch.optim as optim

from advertorch.utils import calc_l2distsq
from advertorch.utils import tanh_rescale
from advertorch.utils import torch_arctanh
from advertorch.utils import clamp
from advertorch.utils import to_one_hot
from advertorch.utils import replicate_input

from .base import Attack
from .base import LabelMixin
from .utils import is_successful


CARLINI_L2DIST_UPPER = 1e10
CARLINI_COEFF_UPPER = 1e10
INVALID_LABEL = -1
REPEAT_STEP = 10
ONE_MINUS_EPS = 0.999999
UPPER_CHECK = 1e9
PREV_LOSS_INIT = 1e6
TARGET_MULT = 10000.0
NUM_CHECKS = 10


class CarliniWagnerL2Attack(Attack, LabelMixin):
    """
    The Carlini and Wagner L2 Attack, https://arxiv.org/abs/1608.04644

    :param predict: forward pass function.
    :param num_classes: number of clasess.
    :param confidence: confidence of the adversarial examples.
    :param targeted: if the attack is targeted.
    :param learning_rate: the learning rate for the attack algorithm
    :param binary_search_steps: number of binary search times to find the
        optimum
    :param max_iterations: the maximum number of iterations
    :param abort_early: if set to true, abort early if getting stuck in local
        min
    :param initial_const: initial value of the constant c
    :param clip_min: mininum value per input dimension.
    :param clip_max: maximum value per input dimension.
    :param loss_fn: loss function
    """

    def __init__(self, predict, num_classes, confidence=0,
                 targeted=False, learning_rate=0.01,
                 binary_search_steps=9, max_iterations=10000,
                 abort_early=True, initial_const=1e-3,
                 clip_min=0., clip_max=1., loss_fn=None):
        """Carlini Wagner L2 Attack implementation in pytorch."""
        if loss_fn is not None:
            import warnings
            warnings.warn(
                "This Attack currently do not support a different loss"
                " function other than the default. Setting loss_fn manually"
                " is not effective."
            )

        loss_fn = None

        super(CarliniWagnerL2Attack, self).__init__(
            predict, loss_fn, clip_min, clip_max)

        self.learning_rate = learning_rate
        self.max_iterations = max_iterations
        self.binary_search_steps = binary_search_steps
        self.abort_early = abort_early
        self.confidence = confidence
        self.initial_const = initial_const
        self.num_classes = num_classes
        # The last iteration (if we run many steps) repeat the search once.
        self.repeat = binary_search_steps >= REPEAT_STEP
        self.targeted = targeted

    def _loss_fn(self, output, y_onehot, l2distsq, const):
        # TODO: move this out of the class and make this the default loss_fn
        #   after having targeted tests implemented
        real = (y_onehot * output).sum(dim=1)

        # TODO: make loss modular, write a loss class
        other = ((1.0 - y_onehot) * output - (y_onehot * TARGET_MULT)
                 ).max(1)[0]
        # - (y_onehot * TARGET_MULT) is for the true label not to be selected

        if self.targeted:
            loss1 = clamp(other - real + self.confidence, min=0.)
        else:
            loss1 = clamp(real - other + self.confidence, min=0.)
        loss2 = (l2distsq).sum()
        loss1 = torch.sum(const * loss1)
        loss = loss1 + loss2
        return loss

    def _is_successful(self, output, label, is_logits):
        # determine success, see if confidence-adjusted logits give the right
        #   label

        if is_logits:
            output = output.detach().clone()
            if self.targeted:
                output[torch.arange(len(label)).long(),
                       label] -= self.confidence
            else:
                output[torch.arange(len(label)).long(),
                       label] += self.confidence
            pred = torch.argmax(output, dim=1)
        else:
            pred = output
            if pred == INVALID_LABEL:
                return pred.new_zeros(pred.shape).byte()

        return is_successful(pred, label, self.targeted)


    def _forward_and_update_delta(
            self, optimizer, x_atanh, delta, y_onehot, loss_coeffs):

        optimizer.zero_grad()
        adv = tanh_rescale(delta + x_atanh, self.clip_min, self.clip_max)
        transimgs_rescale = tanh_rescale(x_atanh, self.clip_min, self.clip_max)
        output = self.predict(adv)
        l2distsq = calc_l2distsq(adv, transimgs_rescale)
        loss = self._loss_fn(output, y_onehot, l2distsq, loss_coeffs)
        loss.backward()
        optimizer.step()

        return loss.item(), l2distsq.data, output.data, adv.data


    def _get_arctanh_x(self, x):
        result = clamp((x - self.clip_min) / (self.clip_max - self.clip_min),
                       min=0., max=1.) * 2 - 1
        return torch_arctanh(result * ONE_MINUS_EPS)

    def _update_if_smaller_dist_succeed(
            self, adv_img, labs, output, l2distsq, batch_size,
            cur_l2distsqs, cur_labels,
            final_l2distsqs, final_labels, final_advs):

        target_label = labs
        output_logits = output
        _, output_label = torch.max(output_logits, 1)

        mask = (l2distsq < cur_l2distsqs) & self._is_successful(
            output_logits, target_label, True)

        cur_l2distsqs[mask] = l2distsq[mask]  # redundant
        cur_labels[mask] = output_label[mask]

        mask = (l2distsq < final_l2distsqs) & self._is_successful(
            output_logits, target_label, True)
        final_l2distsqs[mask] = l2distsq[mask]
        final_labels[mask] = output_label[mask]
        final_advs[mask] = adv_img[mask]

    def _update_loss_coeffs(
            self, labs, cur_labels, batch_size, loss_coeffs,
            coeff_upper_bound, coeff_lower_bound):

        # TODO: remove for loop, not significant, since only called during each
        # binary search step
        for ii in range(batch_size):
            cur_labels[ii] = int(cur_labels[ii])
            if self._is_successful(cur_labels[ii], labs[ii], False):
                coeff_upper_bound[ii] = min(
                    coeff_upper_bound[ii], loss_coeffs[ii])

                if coeff_upper_bound[ii] < UPPER_CHECK:
                    loss_coeffs[ii] = (
                        coeff_lower_bound[ii] + coeff_upper_bound[ii]) / 2
            else:
                coeff_lower_bound[ii] = max(
                    coeff_lower_bound[ii], loss_coeffs[ii])
                if coeff_upper_bound[ii] < UPPER_CHECK:
                    loss_coeffs[ii] = (
                        coeff_lower_bound[ii] + coeff_upper_bound[ii]) / 2
                else:
                    loss_coeffs[ii] *= 10


    def perturb(self, x, y=None):
        x, y = self._verify_and_process_inputs(x, y)

        # Initialization
        if y is None:
            y = self._get_predicted_label(x)
        x = replicate_input(x)
        batch_size = len(x)
        coeff_lower_bound = x.new_zeros(batch_size)
        coeff_upper_bound = x.new_ones(batch_size) * CARLINI_COEFF_UPPER
        loss_coeffs = torch.ones_like(y).float() * self.initial_const
        final_l2distsqs = [CARLINI_L2DIST_UPPER] * batch_size
        final_labels = [INVALID_LABEL] * batch_size
        final_advs = x
        x_atanh = self._get_arctanh_x(x)
        y_onehot = to_one_hot(y, self.num_classes).float()

        final_l2distsqs = torch.FloatTensor(final_l2distsqs).to(x.device)
        final_labels = torch.LongTensor(final_labels).to(x.device)

        # Start binary search
        for outer_step in range(self.binary_search_steps):
            delta = nn.Parameter(torch.zeros_like(x))
            optimizer = optim.Adam([delta], lr=self.learning_rate)
            cur_l2distsqs = [CARLINI_L2DIST_UPPER] * batch_size
            cur_labels = [INVALID_LABEL] * batch_size
            cur_l2distsqs = torch.FloatTensor(cur_l2distsqs).to(x.device)
            cur_labels = torch.LongTensor(cur_labels).to(x.device)
            prevloss = PREV_LOSS_INIT

            if (self.repeat and outer_step == (self.binary_search_steps - 1)):
                loss_coeffs = coeff_upper_bound
            for ii in range(self.max_iterations):
                loss, l2distsq, output, adv_img = \
                    self._forward_and_update_delta(
                        optimizer, x_atanh, delta, y_onehot, loss_coeffs)
                if self.abort_early:
                    if ii % (self.max_iterations // NUM_CHECKS or 1) == 0:
                        if loss > prevloss * ONE_MINUS_EPS:
                            break
                        prevloss = loss

                self._update_if_smaller_dist_succeed(
                    adv_img, y, output, l2distsq, batch_size,
                    cur_l2distsqs, cur_labels,
                    final_l2distsqs, final_labels, final_advs)

            self._update_loss_coeffs(
                y, cur_labels, batch_size,
                loss_coeffs, coeff_upper_bound, coeff_lower_bound)

        return final_advs



from advertorch.attacks import L2PGDAttack
from advertorch.attacks import LinfPGDAttack
from advertorch_examples.utils import TRAINED_MODEL_PATH
import os
import torch
import torch.nn as nn

from advertorch.utils import predict_from_logits
from advertorch.loss import SoftLogitMarginLoss
from advertorch_examples.utils import get_mnist_test_loader


class LeNet5(nn.Module):

    def __init__(self, offset=0., minus_max=False):
        super(LeNet5, self).__init__()
        self.offset = offset
        self.minus_max = minus_max
        self.conv1 = nn.Conv2d(1, 32, kernel_size=3, padding=1, stride=1)
        self.relu1 = nn.ReLU(inplace=True)
        self.maxpool1 = nn.MaxPool2d(2)
        self.conv2 = nn.Conv2d(32, 64, kernel_size=3, padding=1, stride=1)
        self.relu2 = nn.ReLU(inplace=True)
        self.maxpool2 = nn.MaxPool2d(2)
        self.linear1 = nn.Linear(7 * 7 * 64, 200)
        self.relu3 = nn.ReLU(inplace=True)
        self.linear2 = nn.Linear(200, 10)

    def forward(self, x):
        out = self.maxpool1(self.relu1(self.conv1(x)))
        out = self.maxpool2(self.relu2(self.conv2(out)))
        out = out.view(out.size(0), -1)
        out = self.relu3(self.linear1(out))
        out = self.linear2(out)
        out = out * 5
        out = out + self.offset
        if self.minus_max:
            out = out - torch.max(out, dim=1, keepdim=True)[0]
        return out


class ConvNet(nn.Module):
    def __init__(
            self, num_classes=10, return_intermediate=False,
            offset=0., minus_max=False):
        super(ConvNet, self).__init__()
        self.return_intermediate = return_intermediate
        self.minus_max = minus_max
        self.offset = offset

        self.layer1 = nn.Sequential(
            nn.Conv2d(1, 64, kernel_size=3),
            nn.ReLU(),
            nn.Conv2d(64, 64, kernel_size=3),
            nn.ReLU(),
            nn.MaxPool2d((2, 2))
        )

        self.layer2 = nn.Sequential(
            nn.Conv2d(64, 128, kernel_size=3),
            nn.ReLU(),
            nn.Conv2d(128, 128, kernel_size=3),
            nn.ReLU(),
            nn.MaxPool2d(2)
        )

        self.layer3 = nn.Sequential(
            nn.Linear(128 * 4 * 4, 256),
            nn.ReLU(),
        )

        self.layer4 = nn.Sequential(
            nn.Linear(256, 256),
            nn.ReLU(),
            nn.Dropout(0.5)
        )

        self.layer5 = nn.Sequential(
            nn.Linear(256, num_classes)
        )

        self.layers = [self.layer1, self.layer2, self.layer3, self.layer4]

    def forward(self, x):
        relus = list()

        output = self.layer1(x)
        relus.append(output)

        output = self.layer2(output)
        output = output.view(-1, 128 * 4 * 4)
        relus.append(output)

        output = self.layer3(output)
        relus.append(output)

        output = self.layer4(output)
        relus.append(output)

        logits = self.layer5(output)

        # import ipdb
        # ipdb.set_trace()

        # logits = logits / 5.
        # print(torch.max(torch.max(logits, dim=1)[0]
        #                 - torch.min(logits, dim=1)[0]))

        logits = logits + self.offset

        if self.minus_max:
            logits = logits - torch.max(logits, dim=1, keepdim=True)[0]

        if self.return_intermediate:
            return logits, relus
        else:
            return logits


if __name__ == '__main__':
    use_cuda = torch.cuda.is_available()
    device = torch.device("cuda" if use_cuda else "cpu")

    # loss_fn = nn.CrossEntropyLoss(reduction="sum")
    loss_fn = SoftLogitMarginLoss(reduction="sum")


    filename = "mnist_lenet5_clntrained.pt"
    model = LeNet5()
    model.load_state_dict(
        torch.load(os.path.join(TRAINED_MODEL_PATH, filename)))
    model.to(device)
    model.eval()

    # filename = "model_best.pth.tar"
    # model = ConvNet()
    # model.load_state_dict(torch.load(filename)['state_dict'])
    # model.to(device)
    # model.eval()

    situations = [
        (-1000000000000, False),
        (-10000000000, False),
        (-100000000, False),
        (-1000000, False),
        (-1000, False),
        (-40, False),
        (0, False),
        (40, False),
        (1000, False),
        (1000000, False),
        (100000000, False),
        (10000000000, False),
        (1000000000000, False),
        (-1000000000000, True),
        (-10000000000, True),
        (-100000000, True),
        (-1000000, True),
        (-1000, True),
        (-40, True),
        (0, True),
        (40, True),
        (1000, True),
        (1000000, True),
        (100000000, True),
        (10000000000, True),
        (1000000000000, True),
    ]


    batch_size = 100
    loader = get_mnist_test_loader(batch_size=batch_size)
    for cln_data, true_label in loader:
        break
    cln_data, true_label = cln_data.to(device), true_label.to(device)


    logits = model(cln_data)
    torch.save((model, cln_data, true_label, logits), "logits.pt")

    # raise


    # for eps in [0.1, 0.2, 0.3]:
    for eps in [0.3]:
        # eps_iter = eps / 4
        # nb_iter = 100
        eps_iter = eps
        nb_iter = 1
        adversary = LinfPGDAttack(
            model, loss_fn=loss_fn, eps=eps,
            nb_iter=nb_iter, eps_iter=eps_iter,
            rand_init=False, clip_min=0.0, clip_max=1.0,
            targeted=False)

        for offset, minus_max in situations:
            model.offset = offset
            model.minus_max = minus_max
            adv = adversary.perturb(cln_data, true_label)

            # print(torch.norm(adv - cln_data))
            # print((adv != cln_data).sum().item())

            model.offset = 0.
            model.minus_max = False
            advpred = predict_from_logits(model(adv))
            print(
                "eps: {}, offset: {}, minus_max: {}, acc: {:.2f}".format(
                    eps, offset, minus_max,
                    (advpred == true_label).float().mean().item()
                ))


    # eps = 2.
    # eps_iter = eps / 4
    # nb_iter = 100
    # adversary = L2PGDAttack(
    #     model, loss_fn=nn.CrossEntropyLoss(reduction="sum"), eps=eps,
    #     nb_iter=nb_iter, eps_iter=eps_iter,
    #     rand_init=False, clip_min=0.0, clip_max=1.0,
    #     targeted=False)



# https://www.google.com/search?q=pytorch+numerically+stable&oq=pytorch+numerically+stable&aqs=chrome..69i57.6799j0j7&sourceid=chrome&ie=UTF-8

# https://github.com/pytorch/pytorch/issues/1546

# https://discuss.pytorch.org/t/branching-for-numerical-stability/15763

# https://github.com/kleincup/DEEPSEC/issues/3

# https://github.com/BorealisAI/advertorch/issues/63



# Copyright (c) 2018-present, Royal Bank of Canada.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.
#

import warnings

import numpy as np
import torch
import torchvision.transforms.functional as F

from advertorch.utils import torch_allclose
from advertorch.utils import clamp
from advertorch.utils import CIFAR10_MEAN
from advertorch.utils import CIFAR10_STD
from advertorch.utils import MNIST_MEAN
from advertorch.utils import MNIST_STD
from advertorch.utils import NormalizeByChannelMeanStd
from advertorch.utils import PerImageStandardize
from advertorch.utils import torch_flip
from advertorch_examples.utils import bchw2bhwc
from advertorch_examples.utils import bhwc2bchw


def test_mnist_normalize():
    # MNIST
    tensor = torch.rand((16, 1, 28, 28))
    normalize = NormalizeByChannelMeanStd(MNIST_MEAN, MNIST_STD)

    assert torch_allclose(
        torch.stack([F.normalize(t, MNIST_MEAN, MNIST_STD)
                     for t in tensor.clone()]),
        normalize(tensor))


def test_cifar10_normalize():
    # CIFAR10
    tensor = torch.rand((16, 3, 32, 32))
    normalize = NormalizeByChannelMeanStd(CIFAR10_MEAN, CIFAR10_STD)

    assert torch_allclose(
        torch.stack([F.normalize(t, CIFAR10_MEAN, CIFAR10_STD)
                     for t in tensor.clone()]),
        normalize(tensor))


def test_grad_through_normalize():
    tensor = torch.rand((2, 1, 28, 28))
    tensor.requires_grad_()
    mean = torch.tensor((0.,))
    std = torch.tensor((1.,))
    normalize = NormalizeByChannelMeanStd(mean, std)

    loss = (normalize(tensor) ** 2).sum()
    loss.backward()

    assert torch_allclose(2 * tensor, tensor.grad)


def _run_tf_per_image_standardization(imgs):
    import tensorflow as tf
    import tensorflow.image  # noqa: F401

    imgs = bchw2bhwc(imgs)
    placeholder = tf.placeholder(tf.float32, shape=imgs.shape)
    var_scaled = tf.map_fn(
        lambda img: tf.image.per_image_standardization(img), placeholder)

    with tf.Session() as sess:
        tf_scaled = sess.run(var_scaled, feed_dict={placeholder: imgs})
    return bhwc2bchw(tf_scaled)


def test_per_image_standardization():
    imgs = np.random.normal(
        scale=1. / (3072 ** 0.5), size=(10, 3, 32, 32)).astype(np.float32)
    per_image_standardize = PerImageStandardize()
    pt_scaled = per_image_standardize(torch.tensor(imgs)).numpy()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        tf_scaled = _run_tf_per_image_standardization(imgs)
    assert np.abs(pt_scaled - tf_scaled).max() < 0.001


def test_clamp():

    def _convert_to_float(x):
        return float(x) if x is not None else None

    def _convert_to_batch_tensor(x, data):
        return x * torch.ones_like(data) if x is not None else None

    def _convert_to_single_tensor(x, data):
        return x * torch.ones_like(data[0]) if x is not None else None


    for min, max in [(-1, None), (None, 1), (-1, 1)]:

        data = 3 * torch.randn((11, 12, 13))
        case1 = clamp(data, min, max)
        case2 = clamp(data, _convert_to_float(min), _convert_to_float(max))
        case3 = clamp(data, _convert_to_batch_tensor(min, data),
                      _convert_to_batch_tensor(max, data))
        case4 = clamp(data, _convert_to_single_tensor(min, data),
                      _convert_to_single_tensor(max, data))

        assert torch.all(case1 == case2)
        assert torch.all(case2 == case3)
        assert torch.all(case3 == case4)


def test_flip():
    x = torch.randn(4, 5, 6, 7)
    assert (torch_flip(x, dims=(1, 2)) == torch.flip(x, dims=(1, 2))).all()



import torch



for offset in range(0, 100, 10):
    print(offset, torch.exp(torch.tensor(-25. + offset))
          / torch.exp(torch.tensor(25. + offset)))

print()
for offset in range(0, 100, 10):
    print(offset, torch.exp(torch.tensor(-25. - offset))
          / torch.exp(torch.tensor(25. - offset)))


# x = 1.
# y = 10897393428759273457970009.


# print(torch.tensor(x) / torch.tensor(y))
# for ii in range(20):
#     x /= 10000
#     y /= 10000
#     print(torch.tensor(x) / torch.tensor(y))


# x = 1.
# y = 10897393428759273457970009.



# print()
# print()
# print()

# print(torch.tensor(x) / torch.tensor(y))
# for ii in range(20):
#     x *= 10000
#     y *= 10000
#     print(torch.tensor(x) / torch.tensor(y))
